"""sc-githooks - Checks on Git commits

Copyright (c) 2021 Scott Lau
Portions Copyright (c) 2021 InnoGames GmbH
Portions Copyright (c) 2021 Emre Hasegeli
"""

from githooks.config import config
from githooks.base_check import BaseCheck, Severity
from githooks.git import Commit


class CommitCheck(BaseCheck):
    """Parent class for all single commit checks"""
    commit = None

    def prepare(self, obj):
        new = super(CommitCheck, self).prepare(obj)
        if not new or not isinstance(obj, Commit):
            return new

        new = new.clone()
        new.commit = obj
        return new

    def __str__(self):
        return '{} 位于提交 {}'.format(type(self).__name__, self.commit)


class CheckCommitMessage(CommitCheck):
    """提交消息（包含标题）检查
    检查长度、空行等问题"""

    def get_problems(self):
        for line_id, line in enumerate(self.commit.get_message_lines()):
            if line_id == 0:
                continue
            elif line_id == 1:
                if line.strip():
                    yield Severity.ERROR, '第二行应为空行'
            else:
                if line.startswith('    ') or line.startswith('>'):
                    continue

            if line:
                for problem in self.get_line_problems(line_id + 1, line):
                    yield problem

    def get_line_problems(self, line_number, line):
        # if line.rstrip() != line:
        #     line = line.rstrip()
        #     yield (
        #         Severity.ERROR,
        #         '行 {}: 右边不应该包含空格字符'.format(line_number)
        #     )
        #
        # if line.lstrip() != line:
        #     line = line.lstrip()
        #     yield (
        #         Severity.WARNING,
        #         '行 {}: 左边包含空格字符'.format(line_number)
        #     )
        if len(line) > config.get("commit_check.commit_line_max_length"):
            yield (
                Severity.WARNING,
                '行 {}: 超过 {} 字符'.format(line_number, config.get("commit_check.commit_line_max_length"))
            )


class CheckCommitSummary(CommitCheck):
    """提交标题检查
    检查标题的标记、回退等问题"""
    commit_tags = {
        'BREAKING',
        'BUGFIX',
        'CLEANUP',
        'FEATURE',
        'HOTFIX',
        'MESS',
        'MIGRATE',
        'REFACTORING',
        'REVIEW',
        'SECURITY',
        'STYLE',
        'TASK',
        'TEMP',
        'WIP',
        '!!',
    }

    def get_problems(self):
        tags, rest = self.commit.parse_tags()
        # if rest.startswith('['):
        #     yield Severity.WARNING, '未结束的提交标记'
        # if tags:
        #     for problem in self.get_commit_tag_problems(tags, rest):
        #         yield problem
        #     rest = rest[1:]
        #
        # if rest.startswith('Revert'):
        #     for problem in self.get_revert_commit_problems(rest):
        #         yield problem
        #     return

        for problem in self.get_summary_problems(rest):
            yield problem

    # def get_revert_commit_problems(self, rest):
    #     rest = rest[len('Revert'):]
    #     if not rest.startswith(' "') or not rest.endswith('"'):
    #         yield Severity.WARNING, '回退的提交消息格式不合法'
    #
    # def get_commit_tag_problems(self, tags, rest):
    #     used_tags = []
    #     for tag in tags:
    #         tag_upper = tag.upper()
    #         if tag != tag_upper:
    #             yield (
    #                 Severity.ERROR,
    #                 '提交标记 [{}] 不是大写字符'.format(tag)
    #             )
    #         if tag_upper not in CheckCommitSummary.commit_tags:
    #             yield (
    #                 Severity.WARNING,
    #                 '提交标记 [{}] 不在清单中 {}'.format(
    #                     tag, ', '.join(
    #                         '[{}]'.format(t)
    #                         for t in CheckCommitSummary.commit_tags
    #                     )
    #                 )
    #             )
    #         if tag_upper in used_tags:
    #             yield Severity.ERROR, '重复的提交标记 [{}]'.format(tag)
    #         used_tags.append(tag_upper)
    #
    #     if not rest.startswith(' '):
    #         yield Severity.WARNING, '没有使用空格将提交标记与其他内容分开'

    def get_summary_problems(self, rest):
        if not rest:
            yield Severity.ERROR, '无提交标题'
            return

        rest_len = len(rest)
        if rest_len > config.get("commit_check.commit_line_max_length"):
            yield Severity.ERROR, "提交标题不能超过 {} 个字符".format(config.get("commit_check.commit_line_max_length"))
        elif rest_len > config.get("commit_check.commit_summary_max_length"):
            yield Severity.WARNING, "提交标题超过了 {} 个字符".format(config.get("commit_check.commit_summary_max_length"))

        # if '  ' in rest:
        #     yield Severity.WARNING, '存在多个空格字符'

        category_index = rest[:24].find(': ')
        rest_index = category_index + len(': ')
        if category_index >= 0 and len(rest) > rest_index:
            # for problem in self.get_category_problems(rest[:category_index]):
            #     yield problem
            rest = rest[rest_index:]

        for problem in self.get_title_problems(rest):
            yield problem
    #
    # def get_category_problems(self, category):
    #     if not category[0].isalpha():
    #         yield Severity.WARNING, '提交类型以非字母字符开头'
    #     if category.lower() != category:
    #         yield Severity.WARNING, '提交类型包含大写字符'
    #     if category.rstrip() != category:
    #         yield Severity.WARNING, '提交类型右边包含空格字符'

    def get_title_problems(self, rest):
        if not rest:
            yield Severity.ERROR, '无提交标题'
            return
        #
        # first_letter = rest[0]
        # if not first_letter.isalpha():
        #     yield Severity.WARNING, '提交标题以非字母开始'
        # elif first_letter.upper() != first_letter:
        #     yield Severity.WARNING, '提交标题首字母不是大写'
        #
        # if rest.endswith('.'):
        #     yield Severity.WARNING, "提交标题以'.'字符结尾"
        #
        # first_word = rest.split(' ', 1)[0]
        # if first_word.endswith('ed'):
        #     yield Severity.WARNING, '提交标题使用的过去式'
        # if first_word.endswith('ing'):
        #     yield Severity.WARNING, '提交标题使用的进行时'


class CheckChangedFilePaths(CommitCheck):
    """Check file names and directories on a single commit"""

    def get_problems(self):
        for changed_file in self.commit.get_changed_files():
            extension = changed_file.get_extension()
            if (
                    extension in ('pp', 'py', 'sh') and
                    changed_file.path != changed_file.path.lower()
            ):
                yield Severity.WARNING, '{} 文件名使用了大写字母'.format(changed_file)


class CheckBinaryFiles(CommitCheck):
    """Check whether binary files exists on a single commit"""

    def get_problems(self):
        for binary_file in self.commit.get_binary_files():
            yield Severity.WARNING, '文件 {} 是二进制文件'.format(binary_file)



"""sc-githooks - The base check class

Copyright (c) 2021 Scott Lau
Portions Copyright (c) 2021 InnoGames GmbH
Portions Copyright (c) 2021 Emre Hasegeli
"""
import logging
import datetime as dt


from enum import IntEnum


class CheckState(IntEnum):
    NEW = 0
    CLONED = 1
    DONE = 2
    FAILED = 3


class Severity(IntEnum):
    # The numbers are selected to match the Syslog standard.
    ERROR = 3
    WARNING = 4
    NOTICE = 5
    NOTE = 5
    INFO = 6

    def translate(self):
        if self.__eq__(Severity.ERROR):
            return "错误"
        elif self.__eq__(Severity.WARNING):
            return "警告"
        elif self.__eq__(Severity.NOTICE):
            return "注意"
        elif self.__eq__(Severity.NOTE):
            return "注意"
        elif self.__eq__(Severity.INFO):
            return "信息"
        else:
            return "未知"

    @classmethod
    def split(cls, line):
        """Search the severities in the beginning of the string

        It returns the highest severity when non match.
        """
        for name, severity in cls._member_map_.items():
            if line.upper().startswith(name):
                line = line[len(name):].strip(' :-')
                break
        return severity, line


class BaseCheck:
    """The parent class of all checks

    Checks are expanded to different objects by cloning. The subclasses
    has to override prepare() method to clone the check at appropriate
    stage.
    """
    preferred_checks = []
    state = CheckState.NEW
    ERROR_MSG_PREFIX = "GL-HOOK-ERR:"
    NOW_TIME = dt.datetime.now().strftime('%Y-%m-%d %H:%M:%S')

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            # We expect all of the arguments to be initialized with defaults
            # on the class.
            assert hasattr(type(self), key)
            if value:
                setattr(self, key, value)

    def clone(self):
        new = type(self)(**vars(self))
        new.state = CheckState.CLONED
        return new

    def set_state(self, state):
        assert state > CheckState.CLONED
        self.state = max(self.state, state)

    def prepare(self, obj):
        for check in self.preferred_checks:
            if check.prepare(obj):
                return None
        return self

    def print_problems(self):
        header_printed = False
        for severity, problem in self.evaluate_problems():
            if not header_printed:
                print('{} === {} ==='.format(BaseCheck.ERROR_MSG_PREFIX, self))
                logging.info('{} === {} ===\n'.format(BaseCheck.NOW_TIME, self))
                header_printed = True
            print('{} {}: {}'.format(BaseCheck.ERROR_MSG_PREFIX, severity.translate(), problem))
            logging.info('{} {}: {}\n'.format(BaseCheck.NOW_TIME, severity.translate(), problem))
        # if header_printed:
        #     print('{}'.format(BaseCheck.ERROR_MSG_PREFIX))
        self.set_state(CheckState.DONE)

    def evaluate_problems(self):
        assert self.state == CheckState.CLONED
        for severity, problem in self.get_problems():
            if severity <= Severity.ERROR:
                self.set_state(CheckState.FAILED)
            yield severity, problem

    def __str__(self):
        return type(self).__name__


def prepare_checks(checks, obj, next_checks=None):
    """准备检查对象

    它产生准备就绪的checks。不符合要求的checks准备好了吗？
    下一张check list上会有。
    """
    for check in checks:
        prepared_check = check.prepare(obj)
        if prepared_check:
            cloned = prepared_check.state >= CheckState.CLONED
            assert next_checks is not None or cloned

            if cloned:
                yield prepared_check
            else:
                next_checks.append(prepared_check)
